package org.tah.areaCenter;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;

import com.bretth.osmosis.core.container.v0_5.WayContainer;
import com.bretth.osmosis.core.domain.v0_5.Tag;

public class AreaCenterFilter extends PolygonFilter {
	
	private class ProposedCenterPoint {
		double edgeDistance;
		Point2D point;
	}
				
	private Double getAngle(Point2D from, Point2D through, Point2D to) {

		double from_x = from.getX() - through.getX();
		double from_y = from.getY() - through.getY();

		double to_x = to.getX() - through.getX();
		double to_y = to.getY() - through.getY();

		double from_angle = Math.atan2(from_y, from_x) + Math.PI;
		double to_angle = Math.atan2(to_y, to_x) + Math.PI;
		
		return Math.min(from_angle, to_angle) + ( Math.max(from_angle, to_angle) - Math.min(from_angle, to_angle) ) / 2;
	}
	
	private Point2D getLinePoint(Point2D vertex, double angle) {
		double a = 180 - vertex.getX();
		double b = a * Math.tan(angle);

		return new Point2D.Double(vertex.getX() + a, vertex.getY() + b);
	}
			
	private void findVertexCenterPoints(
			ArrayList<ProposedCenterPoint> centerPoints,
			MultiPolygon polygon, int index) {
		
		Polygon outerPolygon = polygon.getOuter();
		double angle = getAngle(outerPolygon.getPoint(index-1), outerPolygon.getPoint(index), outerPolygon.getPoint(index+1));
		Point2D vertex = outerPolygon.getPoint(index);
		Point2D linePoint = getLinePoint(vertex, angle);
		
		// Convert to y = ax + b
		double a = new Line(vertex, linePoint).getSlope();
		double b = vertex.getY() - a * vertex.getX();
		
		// Get all intersections
		ArrayList<Point2D> intersections = new ArrayList<Point2D>();
		intersections.add(vertex);
		 
		for (int k=-1; k<polygon.getInner().size(); k++) {
			Polygon p;
			if (k == -1) {
				p = polygon.getOuter();
			} else {
				p = polygon.getInner().get(k);
			}
			for (int i=0; i<p.size(); i++) {
				if (k >= 0 || (!p.isEqualPoint(i, index) && !p.isEqualPoint(i, index - 1))) {
					Point2D intersection = p.getLine(i).getLineIntersection(a, b);
					if (intersection != null) {
						intersections.add(intersection);
					}
				}
			}			
		}
		
		// Sort intersections
		Collections.sort(intersections, new Comparator<Point2D>() {
			public int compare(Point2D o1, Point2D o2) {
				double result = o1.getX() - o2.getX();
				if (result == 0) {
					result = o1.getY() - o2.getY();
				}
				return (int)Math.signum(result);
			}
		});
		
		if (intersections.size() % 2 == 1) {
			//FIXME this shouldn't happen, anyway it rarely happens probably because line is too close to vertex
			return;
		}
		
		// Add centerpoints
		for (int i=0; i<intersections.size() / 2; i++) {
			ProposedCenterPoint pcp =  new ProposedCenterPoint();
			pcp.point = new Point2D.Double(
							(intersections.get(i*2).getX() + intersections.get(i*2+1).getX()) / 2,
							(intersections.get(i*2).getY() + intersections.get(i*2+1).getY()) / 2
							);
			double dx = intersections.get(i*2).getX() - pcp.point.getX();
			double dy = intersections.get(i*2).getY() - pcp.point.getY();
			pcp.edgeDistance = Math.sqrt(dx*dx + dy*dy);
			centerPoints.add(pcp);
		}
		
	}
	
	private void findCenterPoints(ArrayList<ProposedCenterPoint> centerPoints, 
			MultiPolygon polygon) {
		for (int i=0; i<polygon.getOuter().size(); i++) {
			findVertexCenterPoints(centerPoints, polygon, i);
		}
	}
	
	private Point2D selectBestCenterPoint(ArrayList<ProposedCenterPoint> centerPoints) {
		double sumX = 0;
		double sumY = 0;
		
		// Get maximum edge distance
		double maxEdgeDistance = 0;
		for (ProposedCenterPoint pcp:centerPoints) {
			if (pcp.edgeDistance > maxEdgeDistance) {
				maxEdgeDistance = pcp.edgeDistance;
			}
		}
		
		// Remove all points which have edge distance < 70% of max
		for (Iterator<ProposedCenterPoint> it = centerPoints.iterator(); it.hasNext(); ) {
			if (it.next().edgeDistance < 0.7 * maxEdgeDistance) {
				it.remove();
			}
		}
		
		for (ProposedCenterPoint point:centerPoints) {
			sumX += point.point.getX();
			sumY += point.point.getY();
		}
		
		Point2D averagePoint = 
			new Point2D.Double(sumX / centerPoints.size(), sumY / centerPoints.size());
		
		double minDistance = Double.MAX_VALUE;
		Point2D bestPoint = null;
		
		for (ProposedCenterPoint pcp:centerPoints) {
			if (pcp.point.distance(averagePoint) < minDistance) {
				bestPoint = pcp.point;
				minDistance = pcp.point.distance(averagePoint);
			}
		}
		
		return bestPoint;
	}
		
	private void findCenterPoint(MultiPolygon polygon, WayContainer way) {
		
		
		
		ArrayList<ProposedCenterPoint> centerPoints = new ArrayList<ProposedCenterPoint>();
		Polygon outer = polygon.getOuter();
		Point2D centerPoint = null;
		
		double area = outer.getArea();
//		double len = outer.getLength();

		if (outer.isConvex() || Math.abs(area / outer.getGranScanConvexEnvelope().getArea()) > 0.8) {
			// Convex or almost convex polygon - use centroid
			centerPoint = outer.getCentroid();
			if (!polygon.isPointInside(centerPoint)) {
				centerPoint = null;
			}
		} 

		if (centerPoint == null) {
			// Bobcare algorithm for other polygons
			findCenterPoints(centerPoints, polygon);
			centerPoint = selectBestCenterPoint(centerPoints);
		}

		if (centerPoint != null) {
			way.getEntity().addTag(new Tag("osmarender:areaCenterLat", Double.toString(centerPoint.getX())));
			way.getEntity().addTag(new Tag("osmarender:areaCenterLon", Double.toString(centerPoint.getY())));
		}
		way.getEntity().addTag(new Tag("osmarender:areaSize", Double.toString(area)));
		
/*		//DEBUG
		if (centerPoint != null) {
			Node node = new Node(id--, new Date(), "", centerPoint.getX(), centerPoint.getY());
			Relation relation = new Relation(id--, new Date(), "");
			relation.addTag(new Tag("type", "label"));
			relation.addMember(new RelationMember(node.getId(), EntityType.Node, "label"));
			relation.addMember(new RelationMember(way.getEntity().getId(), EntityType.Way, "object"));
			sink.process(new NodeContainer(node));
			sink.process(new RelationContainer(relation));
		} */
		
		
		//DEBUG Draw convex envelope around every area
/*			Polygon convexEnvelope = polygon.granScanConvexEnvelope();
		
		Way envelope = new Way(--id, new Date(), "");
		for (int i=0; i<convexEnvelope.size(); i++) {
			Node node = new Node(--id, new Date(), "", convexEnvelope.getX(i), convexEnvelope.getY(i));
			sink.process(new NodeContainer(node));
			WayNode wn = new WayNode(id);
			envelope.addWayNode(wn);
		}
		envelope.addTag(new Tag("landuse", "quarry"));
		sink.process(new WayContainer(envelope)); */
	}
				
	@Override
	protected void processPolygon(MultiPolygon polygon,
			WayContainer way) {
		findCenterPoint(polygon, way);
	}

}
