package org.tah.areaCenter;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Stack;

import com.bretth.osmosis.core.domain.v0_5.Way;
import com.bretth.osmosis.core.domain.v0_5.WayNode;

public class Polygon {

	private ArrayList<Point2D> points = new ArrayList<Point2D>();
	
	public Polygon() {
		
	}
	
	public Polygon(Way way, NodesCoordinates nodesCoordinates) {
		load(way, nodesCoordinates);
	}
	
	public void addPoint(Point2D point) {
		points.add(point);
	}
	
	public Point2D getPoint(int index) {
		return points.get((index + 10 * points.size()) % points.size());
	}
	
	public double getX(int index) {
		return getPoint(index).getX();
	}
	
	public double getY(int index) {
		return getPoint(index).getY();
	}
	
	public Line getLine(int index) {
		return new Line(getX(index), getY(index), getX(index + 1), getY(index + 1));
	}
		
	public int size() {
		return points.size();
	}
	
	public void load(Way way, NodesCoordinates nodesCoordinates) {
		// Add all points
		List<WayNode> list = way.getWayNodeList();
		for (int i=0; i<list.size(); i++) {
			Point2D point = nodesCoordinates.getPoint(list.get(i).getNodeId());
			// It's possible point is missing in osm database
			if (point != null) {
				addPoint(point);
			}
		}
		
		// Remove the last one if it's the same as first
		if (points.size() > 0 && getPoint(size()-1).equals(getPoint(0))) {
			points.remove(size() - 1);
		}
	}
		
	private Point2D findMinPoint() {
		Point2D minPoint = getPoint(0);
		for (int i=0; i<size(); i++) {
			Point2D point = getPoint(i);
			if (point.getY() < minPoint.getY()) {
				minPoint = point;
			} else if (point.getY() == minPoint.getY() && point.getX() < minPoint.getX()) {
				minPoint = point;
			}
		}
		return minPoint;
	}
	
	public Polygon getGranScanConvexEnvelope() {
		// Find most bottom - left point
		final Point2D minPoint = findMinPoint();
		
		// Make list of other points
		ArrayList<Point2D> pts = new ArrayList<Point2D>();
		for (int i=0; i<size(); i++) {
			if (getPoint(i) != minPoint) {
				pts.add(getPoint(i));
			}
		}
		
		// Sort them by polar angle
		Collections.sort(pts, new Comparator<Point2D>() {
			public int compare(Point2D o1, Point2D o2) {
				double angle1 = (o1.getX() - minPoint.getX()) / (o1.getY() - minPoint.getY());
				double angle2 = (o2.getX() - minPoint.getX()) / (o2.getY() - minPoint.getY());
				return (int)Math.signum(angle2 - angle1);
			}
		});
		
		// Filter not convex points
		Stack<Point2D> stack = new Stack<Point2D>();
		stack.push(minPoint);
		stack.push(pts.get(0));
		for (int i=1; i<pts.size(); i++) {
	        while (stack.size() >= 2 && crossProduct(stack.get(stack.size() - 2), stack.peek(), pts.get(i)) <= 0) {
	        	stack.pop();	
	        }
            
	        stack.push(pts.get(i));
		}
		
		// Make polygon
		Polygon polygon = new Polygon();
		for (Point2D p:stack) {
			polygon.addPoint(p);
		}
		
		return polygon;
	}
	
	private double crossProduct(Point2D p1, Point2D p2, Point2D p3) {
		return (p2.getX() - p1.getX())*(p3.getY() - p1.getY()) - (p3.getX() - p1.getX())*(p2.getY() - p1.getY());
	}

	
	//TODO Do not create Line2D just for this
	public boolean isConvex() {
		// Dot product of all edges have the same sign for convex polygon
		double d = getLine(-1).dotProduct(getLine(0));
		for (int i=0; i<size()-1; i++) {
			double d2 = getLine(i).dotProduct(getLine(i+1)); 
			if (d2 * d < 0) {
				return false;
			}
		}
		return true;
	}
	
	
	
	public double getArea() {
		double a = 0;
		
		for (int i=0; i<size(); i++) {
			a += getX(i) * getY(i+1) - getX(i + 1) * getY(i);
		}
		
		return a / 2;
	}
	
	public Point2D getCentroid() {
		double x = 0;
		double y = 0;
		
		for (int i=0; i<size(); i++) {
			double a = (getX(i) * getY(i+1) - getX(i+1)*getY(i));
			x += (getX(i) + getX(i+1)) * a;
			y += (getY(i) + getY(i+1)) * a;
		}
		
		double area = getArea();
		return new Point2D.Double(x / (6 * area), y / (6 * area));
	} 
	
	public boolean isPointInside(Point2D point) {
		
		int intersectionCount = 0;
		
		for (int i=0; i<size(); i++) {
			Point2D intersection = getLine(i).getLineIntersection(0, point.getY());
			if (intersection != null && intersection.getX() > point.getX()) {
				intersectionCount++;
			}
		}
		
		return (intersectionCount % 2 == 1);
	}
			
	public boolean isEqualPoint(int index1, int index2) {
		return (index1 + 100 * size()) % size() == (index2 + 100 * size()) % size();
	}
	
	public double getLength() {
		double len = 0;
		for (int i=0; i<size(); i++) {
			len += getLine(i).getLength();
		}
		return len;
	}

}
